@@ -37,9 +37,14 @@
 		* Does no longer borrow the style from its child... not sure if that is good or not.
 		* Reimplemented button relief style, just to test some silliness. =)
 
-	To do:
-		* Reimplement "default" frame, even though it isn't used in gentoo.
-		* Reimplement button press, even though it isn't used in gentoo.
+	1999-12-20:
+                * Can now pull down the menu when the left button has been held down a certain time.
+                  Just like Netscape, with the difference that the button is still active once you have
+                  pulled down the menu.
+
+	1999-12-25:
+		* Reimplemented "default" frame, even though it isn't used in gentoo.
+		* Reimplemented "clicked" signal, even though it isn't used in gentoo.
 	
 	This module was written and is maintained by Johan Hanson <johan@tiq.com>
 */
@@ -53,7 +58,7 @@
 #include <gtk/gtkmenuitem.h>
 
 #include "odemilbutton.h"
-#include "../colorutil.h"
+#include "colorutil.h"
 
 #define DEBUG 0
 
@@ -113,11 +118,13 @@ static void od_emil_button_restore_style_recurse (GtkWidget *widget, gpointer du
 */
 
 /* fiddling with menu and menu items */
+static gint od_emil_button_timeout			(gpointer data);
 static void od_emil_button_detacher			(GtkWidget *widget, GtkMenu *menu);
 static void od_emil_button_position			(GtkMenu *menu, gint *x, gint *y, gpointer userdata);
+static void od_emil_button_position_below	(GtkMenu *menu, gint *x, gint *y, gpointer userdata);
 static void od_emil_button_deactivate		(GtkMenuShell *menu_shell, ODEmilButton *button);
 static void od_emil_button_select			(GtkMenuItem *menu_item, ODEmilButtonChoice *choice);
-static void od_emil_button_popup			(ODEmilButton *button, GdkEvent *event);
+static void od_emil_button_popup			(ODEmilButton *button, guint mb, guint32 time);
 static void od_emil_button_set				(ODEmilButton *button, guint index);
 
 
@@ -209,6 +216,9 @@ static void od_emil_button_init (ODEmilButton *button) {
 	button->button = 0;
 	button->nth = 1;
 	button->config = 0;
+	button->timeout = 0;
+	button->timeout_enabled = 0;
+	button->popdown = 0;
 
 	gtk_widget_ensure_style(GTK_WIDGET(button));
 	button->oldstyle = gtk_widget_get_style(GTK_WIDGET(button));
@@ -237,6 +247,11 @@ static void od_emil_button_destroy (GtkObject *object) {
 	
 	button = OD_EMIL_BUTTON(object);
 
+	if (button->timeout_enabled) {
+		button->timeout_enabled = 0;
+		gtk_timeout_remove(button->timeout);
+	}
+
 	od_emil_button_remove_widget(button, OD_EMIL_BUTTON_SECONDARY);
 	od_emil_button_remove_widget(button, OD_EMIL_BUTTON_PRIMARY);
 	od_emil_button_remove_menu(button);
@@ -477,6 +492,11 @@ static void od_emil_button_size_request (GtkWidget *widget, GtkRequisition *requ
 							   GTK_WIDGET(widget)->style->klass->xthickness);
 	requisition->height = 2 * (GTK_CONTAINER(widget)->border_width + 1 +
 							   GTK_WIDGET(widget)->style->klass->ythickness);
+
+	if (GTK_WIDGET_CAN_DEFAULT (widget)) {
+		requisition->width  += (GTK_WIDGET (widget)->style->klass->xthickness * 2 + 7);
+		requisition->height += (GTK_WIDGET (widget)->style->klass->ythickness * 2 + 7);
+	}
 	
 	od_emil_button_size_calc(button);
 	
@@ -521,13 +541,20 @@ static void od_emil_button_size_allocate (GtkWidget *widget, GtkAllocation *allo
 			child_allocation.width  = MAX(1, (gint)widget->allocation.width - 2 * (child_allocation.x + border_width));
 			child_allocation.height = MAX(1, (gint)widget->allocation.height - 2 * (child_allocation.y + border_width));
 
+			if (GTK_WIDGET_CAN_DEFAULT(button)) {
+				child_allocation.x += (GTK_WIDGET (widget)->style->klass->xthickness + 4);
+				child_allocation.y += (GTK_WIDGET (widget)->style->klass->ythickness + 4);
+				child_allocation.width  = MAX (1, (gint)child_allocation.width -
+										  (gint)(GTK_WIDGET (widget)->style->klass->xthickness * 2 + 7));
+				child_allocation.height = MAX (1, (gint)child_allocation.height -
+										  (gint)(GTK_WIDGET (widget)->style->klass->xthickness * 2 + 7));
+			}
 		  /*
 			n = child_allocation.width - child->requisition.width;
 			if (button->menu && n <= 12 && child == button->choices[OD_EMIL_BUTTON_PRIMARY].widget) {
 				child_allocation.width -= n/2;
 			}
 		  */
-			
 			gtk_widget_size_allocate (child, &child_allocation);
 		}
 	}
@@ -561,6 +588,13 @@ static void od_emil_button_paint_dogear (GtkWidget *widget, GdkRectangle *area)
 		width = widget->allocation.width - 2 * x;
 		height = widget->allocation.height - 2 * y ;
 
+		if (GTK_WIDGET_CAN_DEFAULT(widget)) {
+			x += widget->style->klass->xthickness + 4;
+			y += widget->style->klass->ythickness + 4;
+			width  -= widget->style->klass->xthickness * 2 + 7;
+			height -= widget->style->klass->ythickness * 2 + 7;
+		}
+
 		if (GTK_WIDGET_HAS_FOCUS (widget)) {
 			x += 1;		y += 1;
 			width -= 2;	height -= 2;
@@ -645,6 +679,13 @@ static void od_emil_button_paint_menutab (GtkWidget *widget, GdkRectangle *area)
 		width = widget->allocation.width - 2 * x;
 		height = widget->allocation.height - 2 * y ;
 
+		if (GTK_WIDGET_CAN_DEFAULT(widget)) {
+			x += widget->style->klass->xthickness + 4;
+			y += widget->style->klass->ythickness + 4;
+			width  -= widget->style->klass->xthickness * 2 + 7;
+			height -= widget->style->klass->ythickness * 2 + 7;
+		}
+
 		if (GTK_WIDGET_HAS_FOCUS (widget)) {
 			x += 1;		y += 1;
 			width -= 2;	height -= 2;
@@ -734,6 +775,25 @@ static void od_emil_button_paint (GtkWidget *widget, GdkRectangle *area) {
 		gdk_window_set_back_pixmap (widget->window, NULL, TRUE);
 		gdk_window_clear_area (widget->window, area->x, area->y, area->width, area->height);
 
+		if (   GTK_WIDGET_HAS_DEFAULT (widget)
+			&& GTK_BUTTON(widget)->relief == GTK_RELIEF_NORMAL
+			&& state_type != GTK_STATE_INSENSITIVE)
+		{
+			gtk_paint_box ( widget->style, widget->window,
+							GTK_STATE_NORMAL, GTK_SHADOW_IN,
+							area, widget, "buttondefault",
+							x, y, width, height);
+		}
+		
+		if (GTK_WIDGET_CAN_DEFAULT (widget)) {
+			x += widget->style->klass->xthickness;
+			y += widget->style->klass->ythickness;
+			width -= 2 * x + 7;
+			height -= 2 * y + 7;
+			x += 4;
+			y += 4;
+		}
+
 		if (GTK_WIDGET_HAS_FOCUS (widget)) {
 			x += 1;
 			y += 1;
@@ -753,7 +813,7 @@ static void od_emil_button_paint (GtkWidget *widget, GdkRectangle *area) {
 		}
 
 		gtk_paint_box (widget->style, widget->window, state_type,
-					   shadow_type, area, widget, "multibutton", x, y, width, height);
+					   shadow_type, area, widget, "emilbutton", x, y, width, height);
 
 		/* paint dog ear */
 		if (button->choices[OD_EMIL_BUTTON_SECONDARY].widget
@@ -777,7 +837,6 @@ static void od_emil_button_paint (GtkWidget *widget, GdkRectangle *area) {
 static void od_emil_button_draw (GtkWidget *widget, GdkRectangle *area) {
 	ODEmilButton *button;
 	GdkRectangle child_area;
-	GdkRectangle tmp_area;
 	
 	g_return_if_fail (widget != NULL);
 	g_return_if_fail (OD_IS_EMIL_BUTTON (widget));
@@ -785,15 +844,10 @@ static void od_emil_button_draw (GtkWidget *widget, GdkRectangle *area) {
 	
 	if (GTK_WIDGET_DRAWABLE (widget)) {
 		button = OD_EMIL_BUTTON (widget);
-		
-		tmp_area = *area;
-		tmp_area.x -= GTK_CONTAINER (button)->border_width;
-		tmp_area.y -= GTK_CONTAINER (button)->border_width;
-		
-		od_emil_button_paint (widget, &tmp_area);
+		od_emil_button_paint (widget, area);
 		
 		if (   GTK_BIN (button)->child
-			&& gtk_widget_intersect (GTK_BIN (button)->child, &tmp_area, &child_area))
+			&& gtk_widget_intersect (GTK_BIN (button)->child, area, &child_area))
 		{
 			gtk_widget_draw (GTK_BIN (button)->child, &child_area);
 		}
@@ -805,6 +859,60 @@ static void od_emil_button_draw (GtkWidget *widget, GdkRectangle *area) {
 
 /* internal routines */
 
+static void od_emil_button_position_below (GtkMenu *menu, gint *xp, gint *yp, gpointer userdata)
+{
+	ODEmilButton *button;
+	gint x,y, width,height;
+	gint scrwidth,scrheight;
+	gint btnx,btny, btnwidth,btnheight;
+	gint yt;
+
+	g_return_if_fail (userdata != NULL);
+	g_return_if_fail (OD_IS_EMIL_BUTTON(userdata));
+	button = OD_EMIL_BUTTON(userdata);
+	g_return_if_fail (GTK_WIDGET(menu) == button->menu);
+
+	width =  GTK_WIDGET(menu)->requisition.width;
+	height = GTK_WIDGET(menu)->requisition.height;
+
+	scrwidth  = gdk_screen_width();
+	scrheight = gdk_screen_height();
+	btnwidth = GTK_WIDGET(button)->allocation.width;
+	btnheight = GTK_WIDGET(button)->allocation.height;
+
+	gdk_window_get_origin(GTK_WIDGET(button)->window, &btnx, &btny);
+	yt = GTK_WIDGET(button)->style->klass->ythickness;
+
+	x = *xp - width/2 - 1;
+	y = btny + btnheight;
+	if (GTK_WIDGET_HAS_FOCUS(button))
+		y--;
+	if (GTK_WIDGET_CAN_DEFAULT(button))
+		y -= yt + 3;
+
+	if (width < btnwidth) {
+		if (x<btnx) x = btnx;
+		else if (x+width > btnx+btnwidth) x = btnx + btnwidth - width;
+	} else {
+		if (x > btnx) x = btnx;
+		else if (x+width < btnx+btnwidth) x = btnx + btnwidth - width;
+	}
+
+	if (y+height > scrheight) {
+		y = btny - height;
+		if (GTK_WIDGET_HAS_FOCUS(button))
+			y++;
+		if (GTK_WIDGET_CAN_DEFAULT(button))
+			y += yt + 4;
+	} else if (y<0)
+		y = 0;
+	if (x+width > scrwidth) x = scrwidth - width;
+	if (x<0) x = 0;
+
+	*xp = x;
+	*yp = y;
+}
+
 static void od_emil_button_position (GtkMenu *menu, gint *xp, gint *yp, gpointer userdata) {
 	ODEmilButton *button;
 	GtkWidget *item, *active = NULL;
@@ -903,16 +1011,44 @@ static void od_emil_button_deactivate (GtkMenuShell *menu_shell, ODEmilButton *b
 	g_return_if_fail (OD_IS_EMIL_BUTTON(button));
 	g_return_if_fail (GTK_WIDGET(menu_shell) == button->menu);
 
+	if (button->popdown) {
+		GdkEventButton	event;
+		GdkWindow		*window;
+		gint			x, y;
+
+		button->popdown = 0;
+		window = GTK_WIDGET(button)->window;
+		gdk_window_get_pointer(window, &x, &y, NULL);
+
+		GTK_BUTTON(button)->in_button = (gdk_window_at_pointer(NULL, NULL)==window);
+		
+		event.type = GDK_BUTTON_RELEASE;
+		event.window = window;
+		event.send_event = 1;
+		event.time = gdk_time_get();
+		event.state = 0;
+		event.button = button->button;
+		event.source = GDK_SOURCE_MOUSE;
+		event.x = x;
+		event.y = y;
+		event.pressure = 0.5;
+		event.xtilt = 0.0;
+		event.ytilt = 0.0;
+		event.source = 0;
+		event.deviceid = 0;
+		
+		gtk_signal_emit_by_name(GTK_OBJECT(button), "button_release_event", &event);
+	}
+
   #if DEBUG
 	puts("od_emil_button_deactivate()");
   #endif
 }
 
 
-static void od_emil_button_popup (ODEmilButton *button, GdkEvent *event) {
+static void od_emil_button_popup (ODEmilButton *button, guint mb, guint32 time) {
 	GtkWidget *item, *widget;
 	GList *list;
-	guint16 mb;
 
   #if DEBUG	
 	puts("enter od_emil_button_popup()");
@@ -921,10 +1057,11 @@ static void od_emil_button_popup (ODEmilButton *button, GdkEvent *event) {
 	if (!button->menu)
 		return;
 
-	/* go */
+	/*
 	mb = 0;
 	if (event->type == GDK_BUTTON_PRESS)
 		mb = ((GdkEventButton *) event)->button;
+	*/
 
 	if (button->oldstyle && button->nth) {
 	  /*
@@ -934,7 +1071,7 @@ static void od_emil_button_popup (ODEmilButton *button, GdkEvent *event) {
 	}
 	
 	gtk_menu_popup (GTK_MENU (button->menu), NULL, NULL, od_emil_button_position, button,
-					mb, ((GdkEventButton *)event)->time);
+				    mb, time);
 }
 
 
@@ -1050,6 +1187,9 @@ static void od_emil_button_leave (GtkButton *button) {
 	g_return_if_fail (OD_IS_EMIL_BUTTON (button));
 	emil_button = OD_EMIL_BUTTON(button);
 
+	if (emil_button->popdown)
+		return;
+
 	if (emil_button->button > 1)
 		od_emil_button_set(OD_EMIL_BUTTON(button), OD_EMIL_BUTTON_PRIMARY);
 
@@ -1086,6 +1226,14 @@ static gint od_emil_button_button_release (GtkWidget *widget, GdkEventButton *ev
 	if (mouse != button->button)
 		return FALSE;
 
+	if (button->timeout_enabled) {
+		button->timeout_enabled = 0;
+		gtk_timeout_remove(button->timeout);
+	}
+
+	if (button->popdown)
+		g_warning ("popdown==TRUE");
+
 	gtk_grab_remove(widget);
 	od_emil_button_set(button, OD_EMIL_BUTTON_PRIMARY);
 
@@ -1105,6 +1253,9 @@ static gint od_emil_button_button_release (GtkWidget *widget, GdkEventButton *ev
 				gtk_widget_draw (widget, NULL);
 			}
 			gtk_signal_emit(GTK_OBJECT(button), emil_button_signals[ACTIVATE], choice->user_data, mouse-1);
+			if (mouse == 1)
+				gtk_signal_emit(GTK_OBJECT(button),
+								GTK_WIDGET_CLASS(GTK_OBJECT(button)->klass)->activate_signal);
 		}
 	} else if (GTK_WIDGET_STATE(widget) != state_type) {
 		gtk_widget_set_state(widget, state_type);
@@ -1139,9 +1290,11 @@ static gint od_emil_button_button_press (GtkWidget *widget, GdkEventButton *even
 	if (button->button && mouse != button->button)
 		return TRUE;
 
-	if (mouse >= 3) {
+	if (   mouse >= 3
+		|| (mouse == 1 && button->choices[OD_EMIL_BUTTON_PRIMARY].widget == NULL))
+	{
 		if (button->menu && !button->config) {
-			od_emil_button_popup(button, (GdkEvent *)event);
+			od_emil_button_popup(button, event->button, event->time);
 			return TRUE;
 		}
 	} else {
@@ -1149,15 +1302,19 @@ static gint od_emil_button_button_press (GtkWidget *widget, GdkEventButton *even
 			if (mouse == 2 && button->choices[1].widget)
 				od_emil_button_set(button, OD_EMIL_BUTTON_SECONDARY );
 		} else {
-			if (   !button->config
-				&& (   (mouse == 1 && button->choices[OD_EMIL_BUTTON_PRIMARY].widget == NULL)
-					|| (mouse != 1 && button->choices[OD_EMIL_BUTTON_SECONDARY].widget == NULL)))
-			{
+			if (mouse == 1 && button->choices[OD_EMIL_BUTTON_PRIMARY].widget == NULL) {
+				if (button->menu && button->nth) {
+					od_emil_button_popup(button, event->button, event->time);
+					return TRUE;
+				}
+				return FALSE;
+			} else if (mouse != 1 && button->choices[OD_EMIL_BUTTON_SECONDARY].widget == NULL) {
 				return FALSE;
 			}
+			
 			od_emil_button_set(button, mouse==1 ? OD_EMIL_BUTTON_PRIMARY : OD_EMIL_BUTTON_SECONDARY );
 		}
-		
+
 		button->button = mouse;
 
 		if (!GTK_WIDGET_HAS_FOCUS (widget))
@@ -1170,11 +1327,52 @@ static gint od_emil_button_button_press (GtkWidget *widget, GdkEventButton *even
 		gtk_grab_add(widget);
 
 		GTK_BUTTON(button)->button_down = 1;
+
+		if (button->menu && !button->timeout_enabled && !button->config && button->nth) {
+			button->timeout_enabled = TRUE;
+			button->timeout = gtk_timeout_add(OD_EMIL_BUTTON_TIMEOUT, od_emil_button_timeout, button);
+		}
+		
 		return TRUE;
 	}
 	return FALSE;
 }
 
+static gint od_emil_button_timeout (gpointer data) {
+	ODEmilButton	*button;
+	guint32			mask;
+
+	GDK_THREADS_ENTER ();
+
+	if (data == NULL)
+		goto die;
+	if (!OD_IS_EMIL_BUTTON(data))
+		goto die;
+	button = OD_EMIL_BUTTON(data);
+	if (button->timeout_enabled) {
+		button->timeout_enabled = 0;
+
+		if (   (GTK_WIDGET(button)->state != GTK_STATE_INSENSITIVE)
+			&& button->menu
+			&& !button->config
+			&& button->button
+			&& button->button < 3
+			&& GTK_BUTTON(button)->in_button)
+		{
+			button->popdown = 1;
+
+			if (button->oldstyle && button->nth)
+				col_set_style (GTK_WIDGET(button->menu), gtk_widget_get_style(GTK_WIDGET(button)));
+
+			gtk_menu_popup (GTK_MENU (button->menu), NULL, NULL, od_emil_button_position_below, button,
+							0, gdk_time_get());
+		}
+	}
+
+  die:
+	GDK_THREADS_LEAVE ();
+	return 0;
+}
 
 static void od_emil_button_detacher (GtkWidget *widget, GtkMenu *menu) {
 	ODEmilButton *button;
